#!/usr/bin/env Rscript

################################################################################
# Coordinate Vegan analytics, plot results
################################################################################

library(vegan)
library(ggplot2)
library(plyr)
library(gplots)
library(optparse)

dataDir <- '.'

################################################################################
# parse command line, also setup defaults for interactive debugging
################################################################################

option_list <-
  list( 
       make_option(c("-u", "--unifrac"), action="store_true", default=FALSE,
                   help="Do unifrac processing [default]"),
       make_option(c("-t", "--topN"), default="3", type="integer",
                   help="Display topN OTUs in biplots"))

if(interactive()) {                     # for debugging only
  arguments <- list(args=c('sampleDesc.csv', 'ecoData.csv'),
                    options=list(unifrac=T, topN=3))
} else {                                # normal argv case
  parser <- OptionParser(usage = "%prog [options] sampleDesc.csv ecoDesc.csv",
                         option_list=option_list)
  arguments <- parse_args(parser, positional_arguments = TRUE)
}

opt <- arguments$options
if(length(arguments$args) != 2) {
  cat("Incorrect number of required positional arguments\n\n")
  print_help(parser)
  stop(1)
} else {
  sampleDescPath <- (arguments$args)[1]
  envDescPath <- (arguments$args)[2]
}

################################################################################
# Read environmental vars and sample descriptions
################################################################################

envDesc <- read.csv(envDescPath, as.is=T, row.names=1)
sampleDesc <- read.csv(sampleDescPath, as.is=T, comment.char = "#")
rownames(sampleDesc) <- sampleDesc$shortSampleName

################################################################################
# unifracDistFile read [un]weighted unifrac distance matrix file
################################################################################

unifracDistFile <- function(weightedStr) {
  file.path(dataDir, "unifrac.out",
            sprintf("dist_%s_unifrac_Qiime_communityTable.txt", weightedStr))
}

################################################################################
# Create a pdf graphics device
################################################################################

openPDF <- function (f) { 
  pdf(file=file.path(dataDir, sprintf('%s.pdf', f)), width=11, height=8.5)
}

################################################################################
# Close a pdf graphics device
################################################################################

closePDF <- function() {
  dev.off()
}

################################################################################
# readOtuTable read the OTU table produced by makeCommunityTable.
################################################################################

readOtuTable <- function() {
  df <- read.csv(file.path(dataDir, 'communityTable.csv'), row.names=1)
  print('Sequences in each sample')
  print(rowSums(df))
  print('OTUs in each sample')
  otuBools <- apply(df, 2, function(c) c>0)
  print(rowSums(otuBools))
  df
}


################################################################################
# seqsPerSample: count the total sequences for each biological sample
# of the OTU table
################################################################################

seqsPerSample <- function (otuTable) {
  rowSums(otuTable)
}


################################################################################
# readUnifracDist: read the UniFrac distance matrix.  Adjust
# constrained Qiime names.
################################################################################

readUnifracDist <- function (fn) {

  ## When QIIME/unifrac writes distance files, the final line lacks a
  ## \n, which causes a warning when we read.csv those files.
  ## Suppress that warning.
  suppressWarnings(dm <- read.csv(fn, sep='\t', row.names=1))

  ## re-rename the sample names from qiime-world to our-world (see
  ## makeOTUtable.R/genQiimeOTUtable()
  stopifnot(rownames(dm)==colnames(dm)) # MB symmetrical
  sampleNames <- gsub('.', '_', rownames(dm), fixed=T)

  ## make sure all remapped names are known
  stopifnot(sampleNames %in% sampleDesc$shortSampleName)  

  rownames(dm) <- colnames(dm) <- sampleNames
  dm
}

################################################################################
# doRarefy:
# return a list with elements [[sampleDesc$group]] = attributes of that plot
################################################################################

doRarefy <- function(otuTable) {
  message('computing vegan/rarefy')
  increment <- 100                      # sequence increment along horizontal axis, adjustable
  seqs <- seqsPerSample(otuTable)  
  fullSamplePoints <- seq(from=1, to=max(seqs), by=increment)


  ## compute rarefaction matrix rareMat [subsampleDepth, plot]
  ## colnames are plotnames
  rareMat <-
    apply(as.matrix(otuTable), 1,
                     function(row) {
                       samples <- fullSamplePoints [fullSamplePoints<=sum(row)]
                       r <- rarefy(row, samples)
                       colnames(r) <- NULL
                       ## create full-length row, fill empty positions with NAs
                       res <- c(r[1,], rep(NA, length(fullSamplePoints)-length(samples)))
                       res
                     })

  raredf <- as.data.frame(rareMat)
  raredf$sample <- fullSamplePoints

  ## Use matplot to draw the set of rarefaction lines
  oldPar <- par(cex.main=1.3, cex.axis=1.3, cex.lab=1.3,
                xpd=NA, oma=c(0,1,0,12)   # allocate space for legend on right
                )
  result <-
    dlply(sampleDesc, .(group),
          function (rp)
          {
            ## The columns in rp are unordered.  Here we reorder two ways:
            ##
            ## 1) reorder by column name, used for legend and assigning line colors
            ##
            ## 2) in matplot, order (decreasing) by curve length, so shorter curves
            ##    are plotted last, thereby increasing readability

            ## Reorder by column name & assign colors
            curveDf <- raredf[,sort(rp$shortSampleName), drop=F]
            colors <- rich.colors(ncol(curveDf))
            names(colors) <- colnames(curveDf)
          
            ## compute plotting order by # points in each curve
            curvePoints <- apply(curveDf, 2, function(col) sum(!is.na(col)))
            curveOrder <- order(curvePoints, decreasing=T)

            ## Determine longest curve to limit x axis
            expandHorizForLegend <- 1.0
            xlims <- c(min(fullSamplePoints),
                       expandHorizForLegend*fullSamplePoints[max(curvePoints)])

            matplot(fullSamplePoints, curveDf[,curveOrder], type='l', lty=1,
                    main=rp$group[1],
                    xlab='Sequences', ylab='OTUs',
                    col=colors[curveOrder],
                    lwd=3,
                    xlim=xlims)

            legend(par("usr")[2],par("usr")[4], # place outside plot, upper-right
                   colnames(curveDf), col=colors, lty=1, lwd=3,
                   bty='n',             #no box
                   cex=1.3,
                   ncol=1)
            result <- list(colors=colors, curveDf=curveDf)
          })
  par(oldPar)
  result
}

################################################################################
# doSpeciesRichness: compute species richness with Vegan
################################################################################

doSpeciesRichness  <- function(otuTable) {
  print ('Vegan/estimateR')
  print (round(estimateR(otuTable), digits=2))
}


################################################################################
# doShannonDiversity: compute Shannon diversity on OTU counts
################################################################################

doShannonDiversity <- function(otuTable) {
  ## For alpha diversity with count-based data, Shannon's and Simpson's indices are
  ## relevant.

  ## See Lozupone2008, Species divergence and the measurement of
  ## microbial diversity, Table 1.
  
  message('Shannon diversity')
  shannonDiversity <- round(diversity(otuTable, index='shannon'), digits=4)
  shannonDF <- data.frame(plot=names(shannonDiversity),
                          shannon=shannonDiversity,
                          seqs=seqsPerSample(otuTable))

  shannonDF$plot <- reorder(shannonDF$plot, shannonDF$shannon,
                            function(x) {
                              stopifnot (length(x) ==1)
                              x})

  p <- ggplot(shannonDF) +
    geom_point(aes(x=shannon, y=plot, size=seqs, colour=seqs)) +
    scale_size (range=c(5,15)) + scale_colour_gradientn (colours=rich.colors(20)) +
    ggtitle("Shannon Diversity\n") +
    guides(colour=guide_legend('seqs'), size=guide_legend('seqs'))
  print(p)
}

################################################################################
# genHeatmap: cluster and produce a heatmap of the biological samples.
# This is used on Bray-Curtis and UniFrac distances.
#
################################################################################

genHeatmap <- function(distMat) {

  ## To disable the color key/histogram, set showKey <- FALSE.  
  ## 
  ## See gplots/heatmap.2 documentation for information on adjusting
  ## placement of the color key using the lmat, lhei, and lwid options.
  showKey <- TRUE

  m <- as.matrix(distMat)
  diag(m) <- NA # remove diagonal to get more resolution from the color scale

  cellNotes <- sub('0.', '.', round(m, 2), fixed=T) 
  dim(cellNotes) <- dim(m)            # sub converts matrix -> vector, this restores 2D
  heatmap.2(m, symm=T, trace='none',  # col=brewer.pal(11,'Spectral'),
            dist=as.dist,
            dendrogram='row',
            key=showKey,
            cellnote=cellNotes, notecex=1.2, notecol='black', na.color='gray',
            margins=c(7,7))
}

################################################################################
# Calculate the Bray-Curtis distance matrix based on the OTU table.
#
# For beta diversity, Lozupone2008 Table 1 shows that with
# species-based count data, Sorensen quantitative index and
# Morisita-Horn measure are used.  Ramette2007 agrees, suggests
# Sorensen or Jaccard.  (Note: Sorensen is same as Bray-Curtis, see
# Cha2007 Comprehensive Survey on Distance Similarity Measures between
# Probability Density Functions.)
# 
# Hellinger transformation (sqrt of proportion) downweights abundant
# species, see Legendre2012, From Classical to Canonical Ordination

################################################################################

calcBrayDist <- function(otuTable) {
  vegdist(decostand(otuTable, method='hellinger'), method='bray', binary=F)  
}

################################################################################
# doPCA computes and plots PCoA on distance matrices.
################################################################################

doPCA <- function(dm) {
  pc <- prcomp(dm)
  pcPercentages <- round(100*(pc$sdev^2) / sum(pc$sdev^2))

  df <- data.frame(pc$x)                # extract scores
  df$Row.names <- rownames(df)
  df <- merge(df, sampleDesc, by.x='Row.names', by.y='shortSampleName')

  xLims <- range(c(df$PC1)) * 1.25
  yLims <- range(c(df$PC2)) * 1.05

  p <- ggplot(df) +
    geom_point (aes(x=PC1, y=PC2, colour=group), size=5) +
    geom_text (aes(x=PC1, y=PC2, label=Row.names), hjust=0.0, vjust=-1) +
    scale_x_continuous (sprintf ("PC1 (%d%%)", pcPercentages[1]), limits=xLims) +
    scale_y_continuous (sprintf ("PC2 (%d%%)", pcPercentages[2]), limits=yLims) +
    # theme(legend.position=c(0.2,0.7), legend.justification=c(0,0)) +
    theme(legend.background=element_rect(colour='gray', fill='gray99', size=0.4, linetype='solid')) +
    theme(axis.title.x=element_text(vjust=0, size=14)) +
    theme(panel.grid.minor=element_blank(), panel.grid.major=element_blank()) +
    theme(plot.margin = unit(c(1, 0.5, 1, 0.5), "lines"))
  print(p)
  # print(p + opts(legend.position = "none"))
}

################################################################################
# Compute and plot RDA and CCA biplots/triplots of the OTU and
# environmental variable tables.
################################################################################

doBiplots <- function (otuTable) {

  ## plotWhat specfies what is to be plotted, see plot.cca for
  ## details.  Modify as needed for a given application.
  plotWhat <- c('sites', 'bp')

  ## subset the otuTable to contain only the entries in envDesc (requirement of rda/ccs)
  int <- intersect(rownames(otuTable), rownames(envDesc))
  otuTableSubset <- otuTable[int,]

  ## Determine topN OTUs by quantity
  otuQtys <- colSums(otuTableSubset)
  ord <- order(otuQtys, decreasing=T)
  topOTUcolNames <- colnames(otuTableSubset)[head(ord, opt$topN)]

  for (f in c('rda', 'cca')) {
    res <- do.call(f, list(otuTableSubset, envDesc, na.action='na.omit'))
    ## print(summary(res))
    explained <- 100*eigenvals(res)/sum(eigenvals(res))

    sc <- scores(res)
    xlim <- range(c(sc$sites[,1], sc$species[,1]))
    ylim <- range(c(sc$sites[,2], sc$species[,2]))

    plot(res, display=plotWhat, xlim=xlim, ylim=ylim,
         main=f,
         xlab=sprintf('PC1 (%.1f%%)', explained[1]),
         ylab=sprintf('PC1 (%.1f%%)', explained[2]))

    ## plot topN OTUs (code derived from Vegan/plot.cca, but that plots all OTUs)
    if (opt$topN > 0) {
      sc <- scores(res)
      text(sc$species[topOTUcolNames, ], topOTUcolNames,
           col = "red", cex = 0.7)
    }
  }
}


################################################################################
# main
################################################################################

theme_set(theme_bw(base_family= "sans", base_size=14))

openPDF ('veganAnalysis')
otuTable <- readOtuTable ()
rarefyResults <- doRarefy (otuTable)
doShannonDiversity (otuTable)  
doSpeciesRichness  (otuTable)  
doBiplots (otuTable)
closePDF()

## Bray distances
openPDF ('Bray-Curtis analysis')
brayDist <- calcBrayDist (otuTable)
genHeatmap(brayDist)
doPCA(brayDist)
closePDF()

## Unifrac distances
if (opt$unifrac) {
  for (wt in c("weighted", "unweighted")) {
    wtStr <- sprintf('%s Unifrac analysis', wt)
    print(paste('processing', wtStr))
    openPDF (wtStr)
    ufDist <- readUnifracDist (unifracDistFile(wt))
    genHeatmap(as.dist(ufDist))
    doPCA (ufDist)
    closePDF()
  }
}
