#!/usr/bin/env Rscript

################################################################################
# Transform the ESPRIT-Table output + clean sequences into an OTU
# table.  Compute representative sequences from each OTU.
################################################################################

library(plyr)
library (Biostrings)

## From esprit-tree doc:
##
## 3.3 Mapping Clustering Result back to Original Data pbpcluster uses
## a trimmed sequence dataset. The command invmap.pl maps the
## clustering result generated by pbpcluster back to the original
## FASTA file: >> perl invmap.pl sequence.Cluster sequence Clean.map
## sequence.org.Cluster where sequence Clean.map is generated by
## preproc. The result is saved in sequence.org.Cluster, where 0
## refers to the first sequence in the original FASTA file
## sequence.fas.

dataDir <- '.'
clusterFile <- 'agg.ORF.org.Clusters'
origFile <- 'agg.ORF.fa'
sampleDescFile <- 'sampleDesc.csv'

sampleDesc <- read.csv(file.path(dataDir, sampleDescFile), as.is=T, comment.char = "#")

################################################################################
# 
# parseLine: parse input line of the form:  "0.050 |124|131 132|"
#
# Returns a list with 2 elements:
#   dist: the clustering distance from ESPRIT-tree (0.05 in the example above)
#   clusters: list of 1-based numeric indices to lines of agg.ORF.fa
#             For example, from the input above: list of 2: c(125), c(132, 133)
#
# 
################################################################################

parseLine <- function (l) {
  ## return: list, 1 element / input line.
  ## each element: a list of clusters (entries that were separated by "|" in the input)
  ##   (as with the input file, the first element represents the distance within the clusters)
  ## each cluster: a numeric vector, zero offset, indices into PAnS1F6R01_control_seq.ORF.fa
  groupedLine <- strsplit (l, '|', fixed=T)[[1]]
  groupsChr <- llply(groupedLine, function (gs) strsplit(gs, ' '))

  dist <- as.numeric(groupsChr[[1]])
  ## construct indices back to the original fasta file entries.  Add
  ## +1 because ESPRIT-Tree is 0-based, but the Biostring indexing is
  ## 1-based.
  groupsNum <- llply (groupsChr[-1], function(gc) 1+as.numeric(gc[[1]]))
  list(dist=dist, clusters=groupsNum)
}

## read input line of the form:  "0.050 |124|131 132|"
line <- readLines(file.path(dataDir, clusterFile))
stopifnot (length(line)==1)             # we expect exactly one input line
cfi <- parseLine(line)

################################################################################
# parseMetadata parses a DNAStringSet representing agg.ORF.fa and
# extracts the fasta metadata of each sequence.
#
# Return value is a data frame containing columns of metadata, with
# each row corresponding an input sequence.
################################################################################

parseMetadata <- function(ss){
  message('parsing preclustered metadata...')
  sn <- ldply(strsplit(ss, '\\|'), identity)
  stopifnot(substr(sn[,1], 1, 5) == 'file:')         #1
  stopifnot(substr(sn[,2], 1, 4) == 'cnt:')          #3
  stopifnot(substr(sn[,3], 1, 9) == 'hmm_name:')     #4
  stopifnot(substr(sn[,4], 1, 6) == 'score:')        #5
  stopifnot(substr(sn[,5], 1, 10) == 'error_num:')   #6

  data.frame(file = substring(sn[,1], 6),
             cnt = as.numeric(substring(sn[,2], 5)),
             hmmFrameScore = as.numeric(substring(sn[,4],7)),
             errorNum = as.numeric(substring(sn[,5],11)))
}

################################################################################
# makeCommunityTable take the result of the clusters returned by
# parseLine, and the agg.ORF.fa metadata, and constructs an OTU count
# matrix [sample, OTU] (as required by Vegan)
#
# Return value is a list:
#    dist: clustering distance from agg.ORF.org.Clusters (e.g., 0.05)
#    communityDF: count matrix [sample, OTU]
################################################################################

makeCommunityTable <- function(cfi, preClustMetadata) {
  plots <- levels(preClustMetadata$file)
  nOTU <- length(cfi$clusters)
  ct <- matrix(as.integer(0), nrow=length(plots), ncol=nOTU)
  rownames(ct) <- plots
  colnames(ct) <- sprintf('OTU%d', 1:nOTU)
  dist <- cfi$dist
  message(paste('computing community table for dist', dist))

  for (plot in plots) {
    for (otu in 1:nOTU) {
      preClustIndices <- cfi$cluster[[otu]]
      md <- preClustMetadata[preClustIndices,]
      ct[plot, otu] <- ct[plot, otu] + sum(md$cnt[md$file==plot])
    }
  }
  list(dist=dist, communityDF=as.data.frame(ct))
}


################################################################################
# pickRepresentative chooses a representative DNA sequence from the
# members of an OTU.  The representative is defined to be the sequence
# that occurs the most frequently in the OTU.
################################################################################

pickRepresentative <- function(cfi, preClustMetadata, preClustSS, outFile) {
  message('pickRepresentative')

  ## find the representative sequence for each OTU (see below for
  ## total ordering transformation).  Create a vector of these
  ## sequences, 1/OTU, in order of OTU numbers
  otus <- sapply(1:length(cfi$clusters),
                 function (otu) {
                   cl <- cfi$clusters[[otu]] 
                   meta <- preClustMetadata[cl,, drop=F]
                   ## i indexes to the sequence representing the OTU group
                   cfi$clusters[[otu]][which.max(meta$cnt * meta$hmmFrameScore)] 
                 }, simplify=T)

  ss <- DNAStringSet(preClustSS[otus])
  names(ss) <- sprintf('%s|clust:OTU%d', names(ss), 1:length(otus))
  writeXStringSet(ss, outFile, format='fasta')
  ss
}

################################################################################
# translateRepresentative transforms a set of DNA OTU representative
# sequences into their amino acid sequences.  The input is assumed to
# contain ORFs without any leading nucleotides, provided by HMM-FRAME
################################################################################

translateRepresentative <- function (DNArepresentative, outFile) {
  AArepresentative <- translate(DNArepresentative)
  names(AArepresentative) <- names(DNArepresentative)
  writeXStringSet(AArepresentative, outFile, format='fasta')
  AArepresentative
}

################################################################################
# genQiimeOTUtable is analogous to makeCommunityTable, except the
# result is transposed per Qiime's input requirement.
################################################################################

genQiimeOTUtable <- function(ct, outFile) {
  ## format OTU table for Qiime convert_biom.py

  ## Qiime sample names can contain only alphanumeric chars and <dot>.
  ## Underscores not allowed see
  ## http://qiime.org/documentation/file_formats.html for more

  qiimeSampleNames <- gsub('_', '.', rownames(ct$communityDF), fixed=T)
  stopifnot(!grepl('[^a-z0-9.]+', qiimeSampleNames, ignore.case=T))

  cat ("#Full OTU Counts\n", file=outFile)
  cat ("#OTU ID", qiimeSampleNames, "taxonomy\n",
       file=outFile, sep='\t', append=T)
  df <- cbind(t(ct$communityDF), "Bacteria") # Bacteria representative placeholder
  write.table(df, file=outFile, append=T, quote=F, sep='\t', row.names=T, col.names=F)
}


################################################################################
# generateConcordance produces the output concordance file; this is a
# fasta file associating each input sequence with its OTU assignment.
################################################################################

generateConcordance <- function(cfi, preClustSS, outFile) {
  message('generating concordance')
  res <- DNAStringSet(preClustSS)
  newnames <- names(res)
  l_ply(1:length(cfi$clusters),
         function (i) {
           ci <- cfi$clusters[[i]]
           otu <- sprintf("clust:OTU%d", i)
           newnames[ci] <<- paste(newnames[ci], otu, sep='|') # lexical scope!!
           })
  names(res) <- newnames
  writeXStringSet(res, outFile, format='fasta')
}

################################################################################
# generateAnnotatedCommunityTable produces a csv report of each OTU.
################################################################################

generateAnnotatedCommunityTable <-
  function (ct, DNArepresentative, AArepresentative, outFile) {
    message ('generating annotated community table')
    df <- as.data.frame(t(ct$communityDF))
    rownames(df) <- colnames(ct$communityDF)
    df$reprDNA <- as.character(DNArepresentative)
    df$reprAA <- as.character(AArepresentative)
    df$repr <- names(DNArepresentative)
  
    write.csv(df, outFile)
    df
  }

preClustSS <- readDNAStringSet(file.path(dataDir, origFile))
preClustMetadata <- parseMetadata (names(preClustSS))

message('processing ...')
ct <- makeCommunityTable(cfi, preClustMetadata)

write.csv(ct$communityDF, file=file.path(dataDir, 'communityTable.csv'))

genQiimeOTUtable (ct, file.path(dataDir, 'Qiime_communityTable.csv'))

generateConcordance(cfi, preClustSS, file.path(dataDir, 'concordance.fasta'))

## summarize the OTU counts of each plot
message ('Summary of OTU counts')
l_ply(rownames(ct$communityDF),
      function (r) {
        message(sprintf('\nOTU counts of %s:',r))
        tbl <- as.data.frame(table(t(ct$communityDF[r,])), optional=T)
        colnames(tbl) <- c('sequence_count', 'number_OTUs')
        print(tbl, row.names=F)
      })

## output OTU clustering representative sequences and translate to AA
DNArepresentative <- pickRepresentative (cfi, preClustMetadata, preClustSS,
                               file.path(dataDir, 'DNA_representative.fasta'))

AArepresentative <- translateRepresentative(DNArepresentative,
                                  file.path(dataDir, 'AA_representative.fasta'))

act <-
  generateAnnotatedCommunityTable (ct, DNArepresentative, AArepresentative,
                                   file.path(dataDir, 'reprAnnotatedCommunityTable.csv'))


