#!/usr/bin/env python

################################################################################
# Coordinate uchime processing.  
#
# FunFrame's agg.ORF.hmmframe.fa has fasta header lines like:
# >file:m3co|cnt:5|hmm_name:Cytochrom_D1|score:157.535|error_num:0|
#
# In de novo mode, Uchime requires fasta header lines like:
# >seq1/ab=5.0/ where ab=<float> represents the abundance.
#
# runUchime.py:
# 1) aggregates identical sequences and their cnt:<int> fields, 
# 2) produces Uchime-compliant input lines, 
# 3) runs Uchime
# 4) parses Uchime output (agg.ORF.hmmframe_uchimeOut.tsv)
# 5) identify non-chimeric sequences
# 6) write out non-chimeric in with FunFrame fasta heades
#
################################################################################


import sys
import os.path
import pdb
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import collections
import re
import subprocess
import csv

reGetCnt = re.compile('cnt:(\d+)')

################################################################################
# nextSeqID/seqgen() is a generator used by consUchimeInputSeqRecord
# to assign seq<nnn> fasta identifiers for Uchime.
################################################################################

def nextSeqID():                # unique sequences for uchime fasta input 
    seqID = 0
    while True:
        yield seqID
        seqID += 1
seqGen = nextSeqID()

################################################################################
# consUchimeInputSeqRecord construct a SeqRecord with an aggregated sequence
################################################################################

def consUchimeInputSeqRecord (seq, inpSeqRecs):
    def parseCnt(inpFAheader):
        return int(reGetCnt.search(inpFAheader).group(1))

    def formatUchimeFAheader (totCnt):
        """format fasta header per UCHIME doc, see de novo mode"""
        return "seq%d/ab=%.1f/" % (seqGen.next(), float(totCnt))

    cnts = [parseCnt(inpSeqRec.id) for inpSeqRec in inpSeqRecs]
    return SeqRecord (Seq(seq), id=formatUchimeFAheader(sum(cnts)), description='')
        

################################################################################
# procesInput (inFn)
#   build dictionaries from the fasta input: 
#        seqDict [seq] -> [FA header, ...]       
# returns seqDict
################################################################################

def processInput(inFn):
    inpSeqs = SeqIO.parse(inFn, "fasta")
    seqDict = collections.defaultdict(list)

    nInputSeqs = 0
    for inpSeq in inpSeqs:
        seqDict[str(inpSeq.seq)].append(inpSeq)
        nInputSeqs += 1

    print ("processInput read %d sequences" % nInputSeqs)
    print ("after dereplication, %d sequences remain" % len(seqDict))
    return seqDict

################################################################################
# runUchime runs the uchime program in an external process
################################################################################

def runUchime(uchimeInFn, uchimeOutFn):
    ## uchime --input seqs.fasta -uchimeout output.uchime
    print('running uchime...')
    exitVal = subprocess.call(
        ['uchime', '--input', uchimeInFn, '--uchimeout', uchimeOutFn, '--quiet'])
    print('uchime exited with %d' % exitVal)
    return exitVal

################################################################################
# processUchimeOutput: parse Uchime output, identify predicted chimeras
################################################################################

def processUchimeOutput(uchimeOutFn):
    """process uchime output, return list of non-chimeras, 
       result as "seq20075/ab=1.0/" fasta headers"""

    def nonChimera(cols):       # return bool if this row is OK
        assert(len(cols) >= 17)
        chimeraFlag=cols[16]
        if chimeraFlag=='N':
            return True
        else:
            assert(chimeraFlag=='Y')
            return False

    tsvReader = csv.reader(open(uchimeOutFn, 'rb'), delimiter='\t')
    return [cols[1] for cols in tsvReader if nonChimera(cols)]

################################################################################
# main
################################################################################

def main():

    if len(sys.argv) != 2:
        print >> sys.stderr, \
           """usage: runUCHIME file.fa 
   writes: file_uchimeIn.fa, file_uchimeOut.tsv, and file_uchimeOut.fa
      file_uchimeIn.fa, file_uchimeOut.tsv are temp files and can be deleted.
      file_uchimeOut.fa is the primary output."""
        sys.exit(1)

    inFn = sys.argv[1]
    inFnRoot, inFnExt = os.path.splitext(inFn)
    uchimeInFn = inFnRoot + "_uchimeIn.fa"
    uchimeOutFn = inFnRoot + "_uchimeOut.tsv"
    outputFn = inFnRoot + "_uchimeOut.fa"

    seqDict = processInput(inFn)

    print ("constructing uchime input sequences")
    uchimeInputSeqRecords = [consUchimeInputSeqRecord (seq, inpSeqRecs) 
                             for seq, inpSeqRecs in seqDict.items()]
    print ("writing uchime input sequences")
    SeqIO.write(uchimeInputSeqRecords, uchimeInFn, 'fasta')
    
    # run Uchime and parse results
    runUchime (uchimeInFn, uchimeOutFn)
    nonChimeraSeqHeaders = processUchimeOutput(uchimeOutFn)

    # associate non-chimeras back to original fasta headers and write out
    uchimeInputSeqDict = SeqIO.to_dict(uchimeInputSeqRecords)

    outputSeqs = []
    for nch in nonChimeraSeqHeaders:
        seqKey = str(uchimeInputSeqDict[nch].seq)
        for sde in seqDict[seqKey]:
            sr = SeqRecord(uchimeInputSeqDict[nch].seq, id=sde.id, description='')
            outputSeqs.append(sr)
            
    print ('writing %d output sequences' % len(outputSeqs))
    SeqIO.write(outputSeqs, outputFn, "fasta")

if __name__ == '__main__':
    main()
